#!/bin/tclsh
# -*- tcl -*- 
#
# The intent of this little (I hope) script is to combine icon files to
# produce one icon file.
#
# run with a list of icon files with last file being the requested destination
#
# A few useful functions...
set max 20
set v 0
proc echo {mess} {
  if {$::v} {
    puts stderr "$mess"
  }
}
proc PutsString { this } {
  upvar $this here
  echo "$this is $here"
}

proc BinaryRead4 { f } {
  binary scan [chan read $f 4] i rtn
  return $rtn
}
proc BinaryRead2 { f } {
  binary scan [chan read $f 2] s rtn
  return $rtn
}
proc BinaryRead1 { f } {
  binary scan [chan read $f 1] c rtn
  return [expr {$rtn & 0xff}]
}

proc BinaryCopy {from floc to tloc size} {
  chan seek $from $floc
  chan seek $to $tloc
  chan copy $from $to -size $size
  chan flush $to
}

proc BinaryWriteHead {to data} {
  # Writes the head info to fileid to
  # The head is 6 bytes long and consists of 3 2-byte fields:
  # Reserved always 0
  # Type always 1 for use (.ico)
  # number of images (data)
  chan seek $to 0
  chan puts -nonewline $to [binary format sss 0 1 $data]
  chan flush $to
}

proc BinaryWriteIconDirEntry {to entNum info dataOffSet} {
  # Writes an Icon directory entry 
  # This code computes where the entry is to go, seeks to there and
  # writes the entry. The entry consists of the icon info and
  # its offSet in the file.
  # The info (name,nobytes) width,1 height,1 colors,1 reserve,1 colorPlanes,2
  # bitsPerPixel,2 size,4
  # the offset is 4 bytes
  # we take the info as a list
  
  chan seek $to [expr {6 + $entNum * 16}]
  chan puts -nonewline $to [binary format ccccssii {*}$info $dataOffSet]
  chan flush $to
}

proc notEnoughParms {{h {}}} {
  if {$h == {}} {
    puts stderr "Not enough arguments:"
  }
   puts stderr "
        makeiconfile.tcl ?-n <no>? ?-v? infile ?infile...? outfile
        where <no> if supplied, is the number of the single icon to put out
        otherwise (-n <no> not given) combine icons in all infile(s) into outfile
        -v turns on verbose mode.

        OR 

        makeiconfile.tcl -h   to get this message
      "
  exit 1
}

proc getFiles {arg} {
  global max icon indx
  # Here we get the list of files and check that each is an icon file.
  # this code builds a list of icons in ::icons 
  # each list entry is: filename iconInfo size fileOffset
  # iconInfo is a list: width height colors res colorplanes bits size
  # size is here twice for ease of use...
  #
  if {[llength $arg] < 2} {notEnoughParms }
  set icon {}
  set indx -1
  set ::outFile [lindex $arg end]
  for {set p 0} {$p < ([llength $arg] -1)} {incr p} {
    set filName [lrange $arg $p $p]
    switch -exact $filName {
      -n -
      --n {
	incr p
	set indx [lindex $arg $p $p]
	if {$p < ([llength $arg] -1)} {continue}
	notEnoughParms
      }
      -h -
      --help { notEnoughParms h }
      -v  {
	set ::v 1
	continue
      }
    }

    if {[file readable $filName]} {
      set r [catch "open $filName r" fid]
      if {$r != 0} {
	puts stderr "could not open $filName, quitting"
	exit 1
      }
      chan config $fid -translation binary
      set id [BinaryRead2 $fid]
      set ty [BinaryRead2 $fid]
      set no [BinaryRead2 $fid]
      if {$id != 0 || $ty != 1 || $no == 0 || $no > $max} {
	puts stderr "Fails header test: id(0) $id, type(1) $ty, no(>0 <$max) $no"
	puts stderr "quitting"
	exit 1
      }
      echo "File $filName : Head id: $id, type: $ty, number of icons: $no"
      for {set im 0} {$im < $no} {incr im} {
	foreach id {width height colors res} {
	  set $id [BinaryRead1 $fid]
	}
	if {$width == 0} {set width 256}
	if {$height == 0} {set height 256}
	foreach id {colorplanes bits} {
	  set $id [BinaryRead2 $fid]
	}
	foreach id {size offset} {
	  set $id [BinaryRead4 $fid]
	}
	foreach id {width height colors res colorplanes bits size offset} {
	  PutsString $id
	}
	lappend icon [list $filName \
			  [list $width $height $colors $res\
			       $colorplanes $bits $size] \
			  $size $offset]
      }
    }
  }
}

proc buildIconFile {outFileName icons {which -1}} {
  if {$icons == {} } {
    puts stderr "No icons found, quitting"
    exit 1
  }
  if {$which != -1 && $which <= [llength $icons]} {
    set icons [lrange $icons $which $which]
  }
  set r [catch "open $outFileName w" toFid]
  if {$r != 0} {
    puts stderr "could not open $outFileName, quitting"
    exit 1
  }
  chan config $toFid -translation binary
  #
  # We will build them in order.  First we scan to figure the addresses
  #
  # The header is 6 bytes and each iconDir entry is 16 bytes..
  #
  set no [llength $icons]
  set adds [expr {16 * $no + 6}] ; # address of first data area
  set last $adds
  foreach icon [lrange $icons 1 end] {
    set last [expr {$last + [lindex $icon 2]}]
    lappend adds $last
  }
  echo  "Offset of icons: $adds"
  # Ok, we can build the file now
  # The header
  BinaryWriteHead $toFid $no
  # The icon dir entries
  set entryNum -1
  foreach icon $icons add $adds {
    BinaryWriteIconDirEntry $toFid [incr entryNum] [lindex $icon 1] $add
  }
  # The icon bits...
  foreach icon $icons add $adds {
    lassign $icon file info size offset
    set fromFid [open $file r]
    chan config $fromFid -translation binary
    BinaryCopy $fromFid $offset $toFid $add $size
  }
  close $toFid
}

#do it
getFiles $argv
buildIconFile $outFile $icon $indx
exit 0
